@@ -8,7 +8,6 @@
 #define CNID_DBD_DB_PARAM_H 1
 
 #include <sys/param.h>
-#include <sys/types.h>
 
 #define DEFAULT_LOGFILE_AUTOREMOVE 1
 #define DEFAULT_CACHESIZE          (8 * 1024) /* KB, so 8 MB */
@@ -34,7 +33,7 @@ struct db_param {
     int max_vols;
 };
 
-struct db_param *db_param_read  (char *);
+extern struct db_param *db_param_read  (char *);
 
 #endif /* CNID_DBD_DB_PARAM_H */
 
