@@ -1,4 +1,5 @@
 /*
+ *
  * (c) 2001 (see COPYING)
  */
 
@@ -10,13 +11,13 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <string.h>
-
 #include <pwd.h>
-#include <atalk/logger.h>
 
+#include <atalk/logger.h>
 #include <atalk/afp.h>
 #include <atalk/uam.h>
 #include <atalk/util.h>
+#include <atalk/compat.h>
 
 #ifndef MIN
 #define MIN(a,b) ((a) < (b) ? (a) : (b))
@@ -52,6 +53,13 @@ static int noauth_login(void *obj, struct passwd **uam_pwd,
 	return( AFPERR_BADUAM );
     }
 
+#ifdef AFS
+    if ( setpag() < 0 ) {
+	LOG(log_error, logtype_uams, "noauth_login: setpag: %s", strerror(errno) );
+	return( AFPERR_BADUAM );
+    }
+#endif /* AFS */
+
     *uam_pwd = pwent;
     return( AFP_OK );
 }
@@ -113,7 +121,7 @@ static int noauth_printer(char *start, char *stop, char *username, struct papfil
 }
 
 
-static int uam_setup(const char *path)
+static int uam_setup(void *handle, const char *path)
 {
   if (uam_register(UAM_SERVER_LOGIN_EXT, path, "No User Authent",
                    noauth_login, NULL, NULL, noauth_login_ext) < 0)
