@@ -1,3 +1,4 @@
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -31,26 +32,25 @@ Netatalk 2001 (c)
 
 #include <atalk/util.h>
 #include <atalk/logger.h>
-
-#define OPEN_LOGS_AS_UID 0
+#include <atalk/unix.h>
 
 #define COUNT_ARRAY(array) (sizeof((array))/sizeof((array)[0]))
 
 #define MAXLOGSIZE 512
 
 #define LOGLEVEL_STRING_IDENTIFIERS { \
-  "LOG_NOTHING",                      \
-  "LOG_SEVERE",                       \
-  "LOG_ERROR",                        \
-  "LOG_WARN",                         \
-  "LOG_NOTE",                         \
-  "LOG_INFO",                         \
-  "LOG_DEBUG",                        \
-  "LOG_DEBUG6",                       \
-  "LOG_DEBUG7",                       \
-  "LOG_DEBUG8",                       \
-  "LOG_DEBUG9",                       \
-  "LOG_MAXDEBUG"}                        
+  "-",                      \
+  "severe",                       \
+  "error",                        \
+  "warn",                         \
+  "note",                         \
+  "info",                         \
+  "debug",                        \
+  "debug6",                       \
+  "debug7",                       \
+  "debug8",                       \
+  "debug9",                       \
+  "maxdebug"}                        
 
 /* these are the string identifiers corresponding to each logtype */
 #define LOGTYPE_STRING_IDENTIFIERS { \
@@ -59,10 +59,11 @@ Netatalk 2001 (c)
   "CNID",                            \
   "AFPDaemon",                       \
   "DSI",                             \
-  "ATalkDaemon",                     \
-  "PAPDaemon",                       \
   "UAMS",                            \
-  "end_of_list_marker"}              \
+  "FCE",                             \
+  "ad",                              \
+  "Spotlight",                       \
+  "end_of_list_marker"}
 
 /* =========================================================================
    Config
@@ -85,23 +86,13 @@ UAM_MODULE_EXPORT logtype_conf_t type_configs[logtype_end_of_list_marker] = {
     DEFAULT_LOG_CONFIG, /* logtype_cnid */
     DEFAULT_LOG_CONFIG, /* logtype_afpd */
     DEFAULT_LOG_CONFIG, /* logtype_dsi */
-    DEFAULT_LOG_CONFIG, /* logtype_atalkd */
-    DEFAULT_LOG_CONFIG, /* logtype_papd */
-    DEFAULT_LOG_CONFIG /* logtype_uams */
+    DEFAULT_LOG_CONFIG, /* logtype_uams */
+    DEFAULT_LOG_CONFIG, /* logtype_fce */
+    DEFAULT_LOG_CONFIG, /* logtype_ad */
+    DEFAULT_LOG_CONFIG  /* logtype_sl */
 };
 
-/* We use this in order to track the last n log messages in order to prevent flooding */
-#define LOG_FLOODING_MINCOUNT 5 /* this controls after how many consecutive messages must be detected
-                                   before we start to hide them */
-#define LOG_FLOODING_MAXCOUNT 1000 /* this controls after how many consecutive messages we force a 
-                                      "repeated x times" message */
-#define LOG_FLOODING_ARRAY_SIZE 3 /* this contols how many messages in flow we track */
-struct log_flood_entry {
-    int count;
-    unsigned int hash;
-};
-static struct log_flood_entry log_flood_array[LOG_FLOODING_ARRAY_SIZE];
-static int log_flood_entries;
+static void syslog_setup(int loglevel, enum logtypes logtype, int display_options, int facility);
 
 /* These are used by the LOG macro to store __FILE__ and __LINE__ */
 static const char *log_src_filename;
@@ -122,140 +113,46 @@ static const unsigned int num_loglevel_strings = COUNT_ARRAY(arr_loglevel_string
    Internal function definitions
    ========================================================================= */
 
-/* Hash a log message */
-static unsigned int hash_message(const char *message)
+static int generate_message(char **message_details_buffer,
+                            char *user_message,
+                            int display_options,
+                            enum loglevels loglevel,
+                            enum logtypes logtype)
 {
-    const char *p = message;
-    unsigned int hash = 0, i = 7;
-
-    while (*p) {
-        hash += *p * i;
-        i++;
-        p++;
-    }
-    return hash;
-}
-
-/*
- * If filename == NULL its for syslog logging, otherwise its for file-logging.
- * "unsetuplog" calls with loglevel == NULL.
- * loglevel == NULL means:
- *    if logtype == default
- *       disable logging
- *    else
- *       set to default logging
- */
-static void setuplog_internal(const char *loglevel, const char *logtype, const char *filename)
-{
-    unsigned int typenum, levelnum;
-
-    /* Parse logtype */
-    for( typenum=0; typenum < num_logtype_strings; typenum++) {
-        if (strcasecmp(logtype, arr_logtype_strings[typenum]) == 0)
-            break;
-    }
-    if (typenum >= num_logtype_strings) {
-        return;
-    }
-
-    /* Parse loglevel */
-    if (loglevel == NULL) {
-        levelnum = 0;
-    } else {
-        for(levelnum=1; levelnum < num_loglevel_strings; levelnum++) {
-            if (strcasecmp(loglevel, arr_loglevel_strings[levelnum]) == 0)
-                break;
-        }
-        if (levelnum >= num_loglevel_strings) {
-            return;
-        }
-    }
-
-    /* is this a syslog setup or a filelog setup ? */
-    if (filename == NULL) {
-        /* must be syslog */
-        syslog_setup(levelnum,
-                     typenum,
-                     logoption_ndelay | logoption_pid,
-                     logfacility_daemon);
-    } else {
-        /* this must be a filelog */
-        log_setup(filename, levelnum, typenum);
-    }
-
-    return;
-}
-
-static void generate_message_details(char *message_details_buffer,
-                                     int message_details_buffer_length,
-                                     int display_options,
-                                     enum loglevels loglevel, enum logtypes logtype)
-{
-    char   *ptr = message_details_buffer;
-    int    templen;
-    int    len = message_details_buffer_length;
+    char *details;
+    int    len;
     struct timeval tv;
     pid_t  pid;
+    char buf[256];
 
-    *ptr = 0;
-
-    /* Print time */
     gettimeofday(&tv, NULL);
-    strftime(ptr, len, "%b %d %H:%M:%S.", localtime(&tv.tv_sec));
-    templen = strlen(ptr);
-    len -= templen;
-    ptr += templen;
-
-    templen = snprintf(ptr, len, "%06u ", (int)tv.tv_usec);
-    if (templen == -1 || templen >= len)
-        return;
-        
-    len -= templen;
-    ptr += templen;
-
-    /* Process name &&  PID */
+    strftime(buf, sizeof(buf), "%b %d %H:%M:%S.", localtime(&tv.tv_sec));
     pid = getpid();
-    templen = snprintf(ptr, len, "%s[%d]", log_config.processname, pid);
-    if (templen == -1 || templen >= len)
-        return;
-    len -= templen;
-    ptr += templen;
-
-    /* Source info ? */
-    if ( ! (display_options & logoption_nsrcinfo)) {
-        char *basename = strrchr(log_src_filename, '/');
-        if (basename)
-            templen = snprintf(ptr, len, " {%s:%d}", basename + 1, log_src_linenumber);
-        else
-            templen = snprintf(ptr, len, " {%s:%d}", log_src_filename, log_src_linenumber);
-        if (templen == -1 || templen >= len)
-            return;
-        len -= templen;
-        ptr += templen;
+    const char *basename = strrchr(log_src_filename, '/');
+    if (basename) {
+        basename++;
+    } else {
+        basename = log_src_filename;
     }
 
-    /* Errorlevel */
-    if (loglevel >= (num_loglevel_chars - 1))
-        templen = snprintf(ptr, len,  " (D%d:", loglevel - 1);
-    else
-        templen = snprintf(ptr, len, " (%c:", arr_loglevel_chars[loglevel]);
 
-    if (templen == -1 || templen >= len)
-        return;
-    len -= templen;
-    ptr += templen;
-
-    /* Errortype */
-    if (logtype<num_logtype_strings) {
-        templen = snprintf(ptr, len, "%s", arr_logtype_strings[logtype]);
-        if (templen == -1 || templen >= len)
-            return;
-        len -= templen;
-        ptr += templen;
+    len = asprintf(&details,
+                   "%s%06u %s[%d] {%s:%d} (%s:%s): %s\n",
+                   buf,
+                   (int)tv.tv_usec,
+                   log_config.processname,
+                   pid,
+                   basename,
+                   log_src_linenumber,
+                   arr_loglevel_strings[loglevel],
+                   arr_logtype_strings[logtype],
+                   user_message);
+    if (len == -1) {
+        *message_details_buffer = "";
+        return -1;
     }
-    
-    strncat(ptr, "): ", len);
-    ptr[len -1] = 0;
+    *message_details_buffer = details;
+    return len;
 }
 
 static int get_syslog_equivalent(enum loglevels loglevel)
@@ -279,11 +176,20 @@ static int get_syslog_equivalent(enum loglevels loglevel)
     }
 }
 
-/* =========================================================================
-   Global function definitions
-   ========================================================================= */
+/* Called by the LOG macro for syslog messages */
+static void make_syslog_entry(enum loglevels loglevel, enum logtypes logtype _U_, char *message)
+{
+    if ( !log_config.syslog_opened ) {
+        openlog(log_config.processname,
+                log_config.syslog_display_options,
+                log_config.syslog_facility);
+        log_config.syslog_opened = true;
+    }
 
-void log_init(void)
+    syslog(get_syslog_equivalent(loglevel), "%s", message);
+}
+
+static void log_init(void)
 {
     syslog_setup(log_info,
                  logtype_default,
@@ -291,10 +197,8 @@ void log_init(void)
                  logfacility_daemon);
 }
 
-void log_setup(const char *filename, enum loglevels loglevel, enum logtypes logtype)
+static void log_setup(const char *filename, enum loglevels loglevel, enum logtypes logtype)
 {
-    uid_t process_uid;
-
     if (loglevel == 0) {
         /* Disable */
         if (type_configs[logtype].set) {
@@ -358,21 +262,11 @@ void log_setup(const char *filename, enum loglevels loglevel, enum logtypes logt
         free(tmp);
 
     } else {
-        process_uid = geteuid();
-        if (process_uid) {
-            if (seteuid(OPEN_LOGS_AS_UID) == -1) {
-                process_uid = 0;
-            }
-        }
+        become_root();
         type_configs[logtype].fd = open(filename,
                                         O_CREAT | O_WRONLY | O_APPEND,
                                         S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
-        if (process_uid) {
-            if (seteuid(process_uid) == -1) {
-                LOG(log_error, logtype_logger, "can't seteuid back %s", strerror(errno));
-                exit(EXITERR_SYS);
-            }
-        }
+        unbecome_root();
     }
 
     /* Check for error opening/creating logfile */
@@ -407,7 +301,7 @@ void log_setup(const char *filename, enum loglevels loglevel, enum logtypes logt
 }
 
 /* Setup syslog logging */
-void syslog_setup(int loglevel, enum logtypes logtype, int display_options, int facility)
+static void syslog_setup(int loglevel, enum logtypes logtype, int display_options, int facility)
 {
     /* 
      * FIXME:
@@ -440,10 +334,60 @@ void syslog_setup(int loglevel, enum logtypes logtype, int display_options, int
         arr_loglevel_strings[loglevel]);
 }
 
-void log_close(void)
+/*
+ * If filename == NULL its for syslog logging, otherwise its for file-logging.
+ * "unsetuplog" calls with loglevel == NULL.
+ * loglevel == NULL means:
+ *    if logtype == default
+ *       disable logging
+ *    else
+ *       set to default logging
+ */
+static void setuplog_internal(const char *loglevel, const char *logtype, const char *filename)
 {
+    unsigned int typenum, levelnum;
+
+    /* Parse logtype */
+    for( typenum=0; typenum < num_logtype_strings; typenum++) {
+        if (strcasecmp(logtype, arr_logtype_strings[typenum]) == 0)
+            break;
+    }
+    if (typenum >= num_logtype_strings) {
+        return;
+    }
+
+    /* Parse loglevel */
+    if (loglevel == NULL) {
+        levelnum = 0;
+    } else {
+        for(levelnum=1; levelnum < num_loglevel_strings; levelnum++) {
+            if (strcasecmp(loglevel, arr_loglevel_strings[levelnum]) == 0)
+                break;
+        }
+        if (levelnum >= num_loglevel_strings) {
+            return;
+        }
+    }
+
+    /* is this a syslog setup or a filelog setup ? */
+    if (filename == NULL) {
+        /* must be syslog */
+        syslog_setup(levelnum,
+                     typenum,
+                     logoption_ndelay | logoption_pid,
+                     logfacility_daemon);
+    } else {
+        /* this must be a filelog */
+        log_setup(filename, levelnum, typenum);
+    }
+
+    return;
 }
 
+/* =========================================================================
+   Global function definitions
+   ========================================================================= */
+
 /* This function sets up the processname */
 void set_processname(const char *processname)
 {
@@ -451,19 +395,6 @@ void set_processname(const char *processname)
     log_config.processname[15] = 0;
 }
 
-/* Called by the LOG macro for syslog messages */
-static void make_syslog_entry(enum loglevels loglevel, enum logtypes logtype _U_, char *message)
-{
-    if ( !log_config.syslog_opened ) {
-        openlog(log_config.processname,
-                log_config.syslog_display_options,
-                log_config.syslog_facility);
-        log_config.syslog_opened = true;
-    }
-
-    syslog(get_syslog_equivalent(loglevel), "%s", message);
-}
-
 /* -------------------------------------------------------------------------
    make_log_entry has 1 main flaws:
    The message in its entirity, must fit into the tempbuffer.
@@ -477,10 +408,8 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
      * are already changed. */
     static int inlog = 0;
     int fd, len;
-    char temp_buffer[MAXLOGSIZE];
-    char log_details_buffer[MAXLOGSIZE];
+    char *user_message, *log_message;
     va_list args;
-    struct iovec iov[2];
 
     if (inlog)
         return;
@@ -495,10 +424,13 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
         if (type_configs[logtype].level >= loglevel) {
             /* Initialise the Messages and send it to syslog */
             va_start(args, message);
-            vsnprintf(temp_buffer, MAXLOGSIZE -1, message, args);
+            len = vasprintf(&user_message, message, args);
             va_end(args);
-            temp_buffer[MAXLOGSIZE -1] = 0;
-            make_syslog_entry(loglevel, logtype, temp_buffer);
+            if (len == -1) {
+                return;
+            }
+            make_syslog_entry(loglevel, logtype, user_message);
+            free(user_message);
         }
         inlog = 0;
         return;
@@ -510,12 +442,13 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
     log_src_linenumber = line;
 
     /* Check if requested logtype is setup */
-    if (type_configs[logtype].set)
+    if (type_configs[logtype].set) {
         /* Yes */
         fd = type_configs[logtype].fd;
-    else
+    } else {
         /* No: use default */
         fd = type_configs[logtype_default].fd;
+    }
 
     if (fd < 0) {
         /* no where to send the output, give up */
@@ -524,149 +457,62 @@ void make_log_entry(enum loglevels loglevel, enum logtypes logtype,
 
     /* Initialise the Messages */
     va_start(args, message);
-    len = vsnprintf(temp_buffer, MAXLOGSIZE -1, message, args);
+    len = vasprintf(&user_message, message, args);
+    if (len == -1) {
+        goto exit;
+    }
     va_end(args);
 
-    /* Append \n */
-    if (len ==-1 || len >= MAXLOGSIZE -1) {
-        /* vsnprintf hit the buffer size*/
-        temp_buffer[MAXLOGSIZE-2] = '\n';
-        temp_buffer[MAXLOGSIZE-1] = 0;
-    }
-    else {
-        temp_buffer[len] = '\n';
-        temp_buffer[len+1] = 0;
-    }
-
-    if (type_configs[logtype].level >= log_debug)
-        goto log; /* bypass flooding checks */
-
-    /* Prevent flooding: hash the message and check if we got the same one recently */
-    int hash = hash_message(temp_buffer) + log_src_linenumber;
-
-    /* Search for the same message by hash */
-    for (int i = log_flood_entries - 1; i >= 0; i--) {
-        if (log_flood_array[i].hash == hash) {
-
-            /* found same message */
-            log_flood_array[i].count++;
-
-            /* Check if that message has reached LOG_FLOODING_MAXCOUNT */
-            if (log_flood_array[i].count >= LOG_FLOODING_MAXCOUNT) {
-                /* yes, log it and remove from array */
-
-                /* reusing log_details_buffer */
-                sprintf(log_details_buffer, "message repeated %i times\n",
-                        LOG_FLOODING_MAXCOUNT - 1);
-                write(fd, log_details_buffer, strlen(log_details_buffer));
-
-                if ((i + 1) == LOG_FLOODING_ARRAY_SIZE) {
-                    /* last array element, just decrement count */
-                    log_flood_entries--;
-                    goto exit;
-                }
-                /* move array elements down */
-                for (int j = i + 1; j != LOG_FLOODING_ARRAY_SIZE ; j++)
-                    log_flood_array[j-1] = log_flood_array[j];
-                log_flood_entries--;
-            }
-
-            if (log_flood_array[i].count < LOG_FLOODING_MINCOUNT)
-                /* log it */
-                goto log;
-            /* discard it */
-            goto exit;
-        } /* if */
-    }  /* for */
-
-    /* No matching message found, add this message to array*/
-    if (log_flood_entries == LOG_FLOODING_ARRAY_SIZE) {
-        /* array is full, discard oldest entry printing "message repeated..." if count > 1 */
-        if (log_flood_array[0].count >= LOG_FLOODING_MINCOUNT) {
-            /* reusing log_details_buffer */
-            sprintf(log_details_buffer, "message repeated %i times\n",
-                    log_flood_array[0].count - LOG_FLOODING_MINCOUNT + 1);
-            write(fd, log_details_buffer, strlen(log_details_buffer));
-        }
-        for (int i = 1; i < LOG_FLOODING_ARRAY_SIZE; i++) {
-            log_flood_array[i-1] = log_flood_array[i];
-        }
-        log_flood_entries--;
-    }
-    log_flood_array[log_flood_entries].count = 1;
-    log_flood_array[log_flood_entries].hash = hash;
-    log_flood_entries++;
-
-log:
-    if ( ! log_config.console) {
-        generate_message_details(log_details_buffer, sizeof(log_details_buffer),
-                                 type_configs[logtype].set ?
-                                     type_configs[logtype].display_options :
-                                     type_configs[logtype_default].display_options,
-                                 loglevel, logtype);
-
-        /* If default wasnt setup its fd is -1 */
-        iov[0].iov_base = log_details_buffer;
-        iov[0].iov_len = strlen(log_details_buffer);
-        iov[1].iov_base = temp_buffer;
-        iov[1].iov_len = strlen(temp_buffer);
-        writev( fd,  iov, 2);
-    } else {
-        write(fd, temp_buffer, strlen(temp_buffer));
+    len = generate_message(&log_message,
+                           user_message,
+                           type_configs[logtype].set ?
+                           type_configs[logtype].display_options :
+                           type_configs[logtype_default].display_options,
+                           loglevel, logtype);
+    if (len == -1) {
+        goto exit;
     }
+    write(fd, log_message, len);
+    free(log_message);
+    free(user_message);
 
 exit:
     inlog = 0;
 }
 
-
-void setuplog(const char *logstr)
+void setuplog(const char *logstr, const char *logfile)
 {
-    char *ptr, *ptrbak, *logtype, *loglevel = NULL, *filename = NULL;
-    ptr = strdup(logstr);
-    ptrbak = ptr;
+    char *ptr, *save;
+    char *logtype, *loglevel;
+    char c;
 
-    /* logtype */
-    logtype = ptr;
+    save = ptr = strdup(logstr);
 
-    /* get loglevel */
-    ptr = strpbrk(ptr, " \t");
-    if (ptr) {
-        *ptr++ = 0;
-        while (*ptr && isspace(*ptr))
-            ptr++;
-        loglevel = ptr;
+    ptr = strtok(ptr, ", ");
 
-        /* get filename */
-        ptr = strpbrk(ptr, " \t");
-        if (ptr) {
-            *ptr++ = 0;
+    while (ptr) {
+        while (*ptr) {
             while (*ptr && isspace(*ptr))
                 ptr++;
+
+            logtype = ptr;
+            ptr = strpbrk(ptr, ":");
+            if (!ptr)
+                break;
+            *ptr = 0;
+
+            ptr++;
+            loglevel = ptr;
+            while (*ptr && !isspace(*ptr))
+                ptr++;
+            c = *ptr;
+            *ptr = 0;
+            setuplog_internal(loglevel, logtype, logfile);
+            *ptr = c;
         }
-        filename = ptr;
-        if (filename && *filename == 0)
-            filename = NULL;
+        ptr = strtok(NULL, ", ");
     }
 
-    /* finally call setuplog, filename can be NULL */
-    setuplog_internal(loglevel, logtype, filename);
-
-    free(ptrbak);
+    free(save);
 }
 
-void unsetuplog(const char *logstr)
-{
-    char *str, *filename;
-
-    str = strdup(logstr);
-
-    /* get filename, can be NULL */
-    strtok(str, " \t");
-    filename = strtok(NULL, " \t");
-
-    /* finally call setuplog, filename can be NULL */
-    setuplog_internal(NULL, str, filename);
-
-    free(str);
-}
